/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mapping;

import java.sql.Connection;

/** Provide access to storage system variables.
 *
 * <p>This interface is passed to generator, notably user generators.</p>
 * @see UserGenerator
 */
public interface StorageContext
{
    /** Return the rank (within its siblings) of the current node being stored.
     * @return a numeric OID
     */
    int getNodeRank();
    
    /** Return the character data attached to the current node if any.
     * @return a String containing concatenation of all character data attached to the
     * current node. Null if node has no value.
     */
    String getNodeData();
    
    /** Return the character data attached to the current node if any, normalized
     * according the XML Schema recommendation.
     * @return a String containing concatenation of all character data attached to the
     * current node. Null if node has no value.
     */
    String getNormalizedNodeData();
    
    /** Return the data attached to the current node if any, converted to a java
     * object.
     * @return a String containing concatenation of all character data attached to the
     * current node. Null if node has no value.
     */
    Object getActualNodeData();
    
    /** Return the Full Qualified Name of the node being stored.
     * @return a string of the form prefix:name if the node being stored is an element or an
     * attribute. Null, otherwise.
     */
    String getQName();
    
    /** Return the local name of the node being stored.
     * @return a String if the node being stored is an element or an attribute. Null, otherwise.
     */
    String getLocalName();
    
    /** Return the namespace URI of the node being stored.
     * @return a String if the node being stored is an element or an attribute. Null, otherwise.
     */
    String getNamespaceURI();
    
    /** Return the character data stored in a tuple waiting to be stored.
     * @param tableIndex index of the table mapping corresponding to tuple
     * containing data.
     * @param columnIndex index of the column in the table mapping used to
     * identify the data.
     * @return an Object.
     */
    Object getRefValue(int tableIndex, int columnIndex);
    
    /** Return the JDBC Connection used by the storage module.
     * <p>This connection may be used, for example, to fetch information from the
     * relational database in order to generate a column value.</p>
     * @return a JDBC connection.
     */
    Connection getConnection();
    
    /** get the document ID set by the user or internally generated.
     * <p><B>This method is only implemented by the Repository.</B></p>
     * @return the document ID. null for the Bridge.
     */
    String getDocumentID();
    
    /** get the numeric ID allocated for the document.
     * <p><B>This method is only implemented by the Repository.</B></p>
     * @return the document ID. -1 for the Bridge.
     */
    long getDocumentOID();

    /** Return the local (to a storage operation) OID (Object IDentifier) of the current
     * node being stored.
     * <p><B>This method is only implemented by the Repository.</B></p>
     * @return a numeric OID. -1 for the Bridge.
     */
    long getOID();
    
    /** Return the universal OID (Object IDentifier) of the current node being stored.
     * This ID is global to the XML Repository.
     * <p><B>This method is only implemented by the Repository.</B></p>
     * @return a numeric OID. -1 for the Bridge.
     */
    long getUOID();
    
    /** Return the path OID (Object IDentifier) of the current node being stored.
     * <p><B>This method is only implemented by the Repository.</B></p>
     * @return a numeric OID. -1 for the Bridge.
     */
    short getPathOID();
}
