/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.repository.dbms;

import java.io.FileOutputStream;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.HashMap;
import java.util.Iterator;

import org.xml.sax.helpers.AttributesImpl;
import org.xquark.repository.util.Functions;
import org.xquark.serialize.XMLSerializer;

/**
 * Class containing information on JDBC to Native SQL types mapping.
 *
 */
public class TypeMap extends HashMap
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name:  $";
    
    public void load(AbstractConnection connection) throws SQLException
    {
        ResultSet rs = connection.getConnection().getMetaData().getTypeInfo();
        while (rs.next())
            addType(new TargetNativeType(rs));
        rs.close();
    }
    
	public void addType(TargetNativeType nativeType)
	{
		TargetNativeType previousType;
		Integer key = new Integer(nativeType.getJDBCType());
		previousType = (TargetNativeType) get(key);
		if (previousType == null)
			put(key, nativeType);
		else // build a chain
			previousType.setNext(nativeType);
	}
	
    public TargetNativeType getNativeType(int JDBCType)
    {
        TargetNativeType targetType = (TargetNativeType)get(new Integer(JDBCType));
        if (targetType == null)
            throw new RuntimeException("The JDBC type " + JDBCType + " is not supported by the database JDBC driver.");
        return targetType;
    }
    
    public String getNativeTypeName(int JDBCType)
    {
        return getNativeType(JDBCType).getNativeTypeName();
    }
    
    public long getSize(int JDBCType)
    {
        return getNativeType(JDBCType).getPrecision();
    }
    
    public boolean isUnsigned(int JDBCType)
    {
        return getNativeType(JDBCType).isUnsigned();
    }
    
    public String getTypeParam(int JDBCType)
    {
        return getNativeType(JDBCType).getTypeParam();
    }
    
    public void discardAllExcept(int JDBCType, String nativeType)
    {
        TargetNativeType type = getNativeType(JDBCType);
        // browse chained list
        while (type != null)
        {
            if (type.getNativeTypeName().equalsIgnoreCase(nativeType))
                break;
            type = type.next;
        }
        // replace the chain by the unique matching type
        if (type != null)
        {
            type.setNext(null);
            put(new Integer(JDBCType), type);
        }
    }
    
    public void serialize(String fileName)
    {
    	try {
    		System.out.println("Starting output...");
    	XMLSerializer serializer = new XMLSerializer(new FileOutputStream(fileName + ".xml"), "UTF-8");
    	serializer.setAutoStreamClose(true);
    	AttributesImpl atts = new AttributesImpl();
    	serializer.startDocument();
    	atts.addAttribute("", "name", "name", "CDATA", fileName);
    	serializer.startElement("","dbmsInfo","dbmsInfo", atts);
    	atts.clear();
    	serializer.startElement("","typeMapping","typeMapping", atts);
    	Iterator it = values().iterator();
    	TargetNativeType wType;
    	String wValue;
    	
    	while (it.hasNext())
    	{
    		wType = (TargetNativeType)it.next();
    		
    		atts.addAttribute("", "jdbcType", "jdbcType", "CDATA", getJDBCTypeName(wType.getJDBCType()));
	    	serializer.startElement("","type","type", atts);
    		atts.clear();
	    	
	    	wValue = wType.getNativeTypeName();
	    	serializer.startElement("","nativeName","nativeName", atts);
	    	serializer.characters(wValue.toCharArray(), 0, wValue.length());
	    	serializer.endElement("","nativeName","nativeName");
    		
	    	wValue = String.valueOf(wType.getPrecision());
	    	serializer.startElement("","precision","precision", atts);
	    	serializer.characters(wValue.toCharArray(), 0, wValue.length());
	    	serializer.endElement("","precision","precision");
    		
	    	wValue = String.valueOf(wType.isUnsigned());
	    	serializer.startElement("","unsigned","unsigned", atts);
	    	serializer.characters(wValue.toCharArray(), 0, wValue.length());
	    	serializer.endElement("","unsigned","unsigned");
    		
	    	wValue = wType.getTypeParam();
	    	if (wValue != null)
	    	{
	    	serializer.startElement("","fixedCreateParams","fixedCreateParams", atts);
	    	serializer.characters(wValue.toCharArray(), 0, wValue.length());
	    	serializer.endElement("","fixedCreateParams","fixedCreateParams");
	    	}
    		
	    	serializer.endElement("","type","type");
    	}
    	
    	serializer.endElement("","typeMapping","typeMapping");
    	serializer.endElement("","dbmsInfo","dbmsInfo");
    	serializer.endDocument();
    	
    	}
    	catch(Exception e)
    	{
    		e.printStackTrace();
    	}
    	
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                  INNER CLASSES
    ///////////////////////////////////////////////////////////////////////////
    public static class TargetNativeType implements Cloneable
    {
        private String nativeTypeName;
        private int JDBCType;   // note : int in java.sql.Types but short in rs
        private long precision;   // note : int in java.sql.Types but short in rs
        private boolean unsigned;
        private String createParam;   // String to concatenate to nativeTypeName when creating table
                                      // (containing precision for instance)
        TargetNativeType next = null;
        
        
		/**
		 * Default constructor for TargetNativeType.
		 */
		public TargetNativeType() {};

        TargetNativeType(ResultSet rs) throws SQLException
        {
            nativeTypeName = rs.getString(1);
            JDBCType = rs.getShort(2);
            precision = rs.getLong(3);  // getInt -> getLong because of Oracle JDBC 8.1.7
            createParam = rs.getString(6);
            unsigned = rs.getBoolean(10);
        }
        
        String getNativeTypeName()
        {
            return nativeTypeName;
        }
        String getTypeParam()
        {
            return createParam;
        }
        int getJDBCType()
        {
            return JDBCType;
        }
        long getPrecision()
        {
            return precision;
        }
        boolean isUnsigned()
        {
            return unsigned;
        }
        public void setNativeTypeName(String typeName)
        {
            nativeTypeName = typeName;
        }
        public void setTypeParam(String param)
        {
            createParam = param;
        }
        public void setPrecision(long prec)
        {
            precision = prec;
        }
        void setNext(TargetNativeType next)
        {
            this.next = next;
        }
        public void setUnsigned(boolean unsigned)
        {
            this.unsigned = unsigned;
        }
        public String toString()
        {
            return "[jdbc=" + JDBCType + ", sql=" + nativeTypeName + ", prec=" + precision
            + ", unsigned=" + unsigned + ", param=" + createParam + "]";
        }
        public Object clone()
        {
            try {
                return super.clone();
            }
            catch (CloneNotSupportedException e) {
                return null;
            }
        }
		/**
		 * Sets the jDBCType.
		 * @param jDBCType The jDBCType to set
		 */
		public void setJDBCType(int jDBCType)
		{
			JDBCType = jDBCType;
		}
		
		public void setJDBCType(String jDBCType)
		{
			JDBCType = Functions.parseJDBCType(jDBCType);
		}

    }
    
	public static String getJDBCTypeName(int type)
	{
		switch (type)
		{
			case Types.ARRAY :
				return "ARRAY";
			case Types.BIGINT :
				return "BIGINT";
			case Types.BINARY :
				return "BINARY";
			case Types.BIT :
				return "BIT";
			case Types.BLOB :
				return "BLOB";
			case Types.CHAR :
				return "CHAR";
			case Types.CLOB :
				return "CLOB";
			case Types.DATE :
				return "DATE";
			case Types.DECIMAL :
				return "DECIMAL";
			case Types.DISTINCT :
				return "DISTINCT";
			case Types.DOUBLE :
				return "DOUBLE";
			case Types.FLOAT :
				return "FLOAT";
			case Types.INTEGER :
				return "INTEGER";
			case Types.JAVA_OBJECT :
				return "JAVA_OBJECT";
			case Types.LONGVARBINARY :
				return "LONGVARBINARY";
			case Types.LONGVARCHAR :
				return "LONGVARCHAR";
			case Types.NULL :
				return "NULL";
			case Types.NUMERIC :
				return "NUMERIC";
			case Types.OTHER :
				return "OTHER";
			case Types.REAL :
				return "REAL";
			case Types.REF :
				return "REF";
			case Types.SMALLINT :
				return "SMALLINT";
			case Types.STRUCT :
				return "STRUCT";
			case Types.TIME :
				return "TIME";
			case Types.TIMESTAMP :
				return "TIMESTAMP";
			case Types.TINYINT :
				return "TINYINT";
			case Types.VARBINARY :
				return "VARBINARY";
			case Types.VARCHAR :
				return "VARCHAR";
			default :
				return "";
		}

	}
}
