/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.repository.mapping;

import java.sql.*;

import org.xquark.repository.dbms.ColumnSpec;
import org.xquark.repository.dbms.TypeMap;

/**
 * This class maintains the meta information about a column in a relational
 * table. This information is useful to verify the integrity constraints
 * in the the relational tables.
 *
 * This class is the default JDBC implementation of ColumnMetaData that uses
 * standard JDBC metadata functions.
 */
public class ColumnMetaDataImpl implements ColumnMetaData
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name:  $";

    /** Whether the column belongs to the primary key */
    protected short primaryKeySeq = -1;

    /** Name of the primary key table when the column is a foreign key */
    protected String refTable = null;

    /** Name of the primary key column when the column is a foreign key */
    protected String refColumn = null;

    /** The name of the table that owns this column */
    protected String tableName;

    /** The name of this column */
    protected String columnName;

    /** The JDBC data type of this column */
    protected short JDBCtype;

    /** The native name of this column's data type */
    protected String nativeType;

    /** The size of this column */
    protected long columnSize;

    /** Number of digit after decimal point */
    protected int decimalDigits;

    /** Whether this column can be null */
    protected boolean nullable;
    
    /** Whether this column is a 'full text' type */
    protected boolean isLongType = false;

    /** Whether this column is a LOB type */
    protected boolean isLOB = false;

    /** Default value */
    protected String columnDef;

    /** The ordinal position of this column within the table's definition */
    protected int ordinalPosition;

    /** Type creation string */
    protected String typeDDLString;
    
    protected ColumnMetaDataImpl(){};
    
    /**
     * Constructor.
     *
     * The constructor of the ColumnMetaData is called by AbstractConnection
     * implementation.
     * @param rs the result set of DatabaseMetadata.getColumns()
     */
    public ColumnMetaDataImpl(ResultSet rs, TypeMap typeMap) throws SQLException
    {
        tableName = rs.getString(3);
        columnName = rs.getString(4);
        JDBCtype = rs.getShort(5);
        nativeType = rs.getString(6);
        columnSize = rs.getInt(7);
        decimalDigits = rs.getInt(9);
        nullable = (rs.getInt(11) == DatabaseMetaData.columnNullable);
        columnDef = rs.getString(13);
        ordinalPosition = rs.getInt(17);
        
        // base type setting
        switch (JDBCtype)
        {
            case Types.TINYINT: case Types.SMALLINT: case Types.INTEGER:
            case Types.BIGINT: case Types.BIT: case Types.NUMERIC:
                JDBCtype = Types.DECIMAL;
                break;
            
            case Types.FLOAT:
                JDBCtype = Types.DOUBLE;
                break;
            
            case Types.CLOB:
                isLOB = true;
                // no break
            case Types.LONGVARCHAR:
                isLongType = true;
                // no break
            case Types.CHAR:
                JDBCtype = Types.VARCHAR;
                break;
                
            case Types.BLOB:
                isLOB = true;
                // no break
            case Types.LONGVARBINARY:
                isLongType = true;
                // no break
            case Types.BINARY: 
                JDBCtype = Types.VARBINARY;
                break;
                
            case Types.DATE: case Types.TIME:
                JDBCtype = Types.TIMESTAMP;
                break;
        }
        
        // type creation string
        typeDDLString = ColumnSpec.getTypeCreationString(
                        getDataType(),
                        typeMap,
                        getColumnSize(),
                        (int)getColumnSize(),
                        getDecimalDigits());
    }

    /**
     * Whether this column can be null.
     */
    public boolean isNullable()
    {
        return nullable;
    }

    public boolean isOptional()
    {
        return isNullable() || getDefaultValue() != null;
    }

    public String getTableName() { return tableName; }

    public String getColumnName() { return columnName; }

    public short getDataType() { return JDBCtype; }

    public String getTypeName() { return nativeType; }

    public long getColumnSize() { return columnSize; }

    public int getDecimalDigits() { return decimalDigits; }

    public String getDefaultValue() { return columnDef; }

    public int getOrdinalPosition() { return ordinalPosition; }

    public boolean isPrimaryKey() { return primaryKeySeq != -1; }

    public short getPrimaryKeySeq() { return primaryKeySeq; }

    public String getRefTable() { return refTable; }

    public String getRefColumn() { return refColumn; }

    public String toString()
    {
        return "#<Column "+getTableName()+"."+getColumnName()+">";
    }

    public void setPrimaryKeySeq(short seq) { primaryKeySeq = seq; }

    public void setRefTable(String name) { refTable = name; }

    public void setRefColumn(String name) { refColumn = name; }
    
    public String getTypeCreationString()
    {
        return typeDDLString;
    }
    
    public boolean isLongType()
    {
        return isLongType;
    }
    
    public boolean isLOB()
    {
        return isLOB;
    }
    
}
