/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.repository.mapping;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.sql.SQLException;
import java.util.*;

import org.xml.sax.*;
import org.xquark.mapping.Generator;
import org.xquark.mapping.UserGenerator;
import org.xquark.repository.dbms.AbstractConnection;
import org.xquark.repository.metadata.RepositoryConstants;
import org.xquark.repository.util.RepositoryProperties;
import org.xquark.schema.*;
import org.xquark.util.DefaultElementHandler;
import org.xquark.util.ElementHandler;
import org.xquark.util.SAXLoader;
import org.xquark.xml.xdbc.XMLErrorHandler;

/** This class is responsible for building the memory reprezentation of the
 * mapping (RepositoryMapping) from a mapping file.
 * @see org.xquark.repository.metadata.RepositoryMapping
 */
public class Loader extends SAXLoader implements MappingConstants
{
    private static final String RCSRevision = "$Revision: 1.6 $";
    private static final String RCSName     = "$Name:  $";

    private RepositoryMapping colMapping =null;
    private MappingFactory factory;
    private XMLReader reader;
    private AbstractConnection conn;
    private SchemaManager manager;
    /** Used to load user generators */
    private ClassLoader classLoader = null;
    private boolean generate =false;
    private String version ="1.0";
    private BuilderMappingContext context =null;
    private Type currentType = null;
    private List scope = new ArrayList();
    /** Type Mappings */
    private HashMap namedMappings = new HashMap();
    /* If not null, warning are sent to it */
    private XMLErrorHandler messageHandler = null;
    /* This flag is used to desactivate OID tables in mapping for mapper */
    private boolean userTablesOnly = false;

    /** Constructor
     * @param reader Parser used to analyse mapping file.
     * @param conn JDBC connection used to retrieve database metadata.
     * @param generate If true, mapping is loaded even if relational tables
     * are not present : mapping is loaded for table generation.
     * If false, table existence must be verified.
     */
    public Loader(
        XMLReader reader,
        AbstractConnection conn,
        SchemaManager manager,
        boolean generate) {
		this(reader, conn, manager, null, generate);
    }

    public Loader(
        XMLReader reader,
        AbstractConnection conn,
        SchemaManager manager,
        ClassLoader loader,
        boolean generate) {
			super(null);
			setElementHandler(new MappingHandler());
			this.reader = reader;
			this.conn = conn;
			this.manager = manager;
			this.generate = generate;
			//    reader.setContentHandler(this);
			if (loader == null)
				classLoader = getClass().getClassLoader();
			else
	        	classLoader = loader;
        }
	
    public RepositoryMapping load(InputSource source, MappingFactory factory, boolean userTablesOnly)
    throws java.io.IOException, SAXException, SQLException
    {
        return load(source, factory, null, userTablesOnly);
    }
            
    public RepositoryMapping load(InputSource source, MappingFactory factory,
            XMLErrorHandler messageHandler, boolean userTablesOnly)
    throws java.io.IOException, SAXException, SQLException
    {
// TO RESTORE !        SchemaTypeInfo.loadTypeMetadata(conn);    // Has to be done somewhere !
        this.colMapping = (RepositoryMapping)factory.getTree();
        this.factory = factory;
        this.messageHandler = messageHandler;
        this.userTablesOnly = userTablesOnly;
        currentType = null;
        scope.clear();
        namedMappings.clear();
        reader.setContentHandler(this);
        reader.parse(source);
        colMapping.initialize();
        return colMapping;
    }

    private void notifyUnknownElement(String uri, String name)
    throws SAXException
    {
        throw new SAXParseException("Syntax error : the element \"" + name
                               + "\" with namespace \"" + uri + "\""
                               + " is not part of the mapping language.",
                               getDocumentLocator());
    }

    private boolean isMappingNamespace(String uri) {
        return MAPPING_URI.equals(uri) || BRIDGE_MAPPING_URI.equals(uri);
    }

    private void initMapping(Attributes atts) throws SAXException
    {
        context = new BuilderMappingContext(factory);   // TO MOVE adding a reset to context ?
        String ver = atts.getValue("", VERSION_ATTR);
        if (ver != null) version = ver;
        String schemaLocations = atts.getValue("", SCHEMA_LOCATION_ATTR);
        if (schemaLocations != null)
        {
            URL base = null;
            String systemId = getDocumentLocator().getSystemId();
            if (systemId != null)
            {
                try
                {
                    base = new URL(systemId);
                } catch (MalformedURLException ex)
                {
                    // should not happen
                }
            }
            StringTokenizer tok = new StringTokenizer(schemaLocations);
            while (tok.hasMoreTokens())
            {
                String ns = tok.nextToken();
                if (!tok.hasMoreTokens())
                {
                    String errMsg = "Odd number of values in schemaLocation attribute: " + schemaLocations;
                    throw new SAXParseException(errMsg, getDocumentLocator());
                }
                String loc = tok.nextToken();
                URL uri = null;
                try
                {
                    if (base == null)
                    {
                        File wDir = new File(".");
                        try
                        {
                            base = new URL(wDir.toURL().toExternalForm());
                        } catch (MalformedURLException ex)
                        {
                            // no op : try without base
                        }
                        uri = new URL(base, loc);
                    }
                    else
                        uri = new URL(base, loc);
                    manager.addSchemaLocation(ns, uri.toString());
                } catch (MalformedURLException ex)
                {
                    throw new SAXParseException("The location '" + loc +
                    "' is malformed or it is relative and your input source does not provide a base URI.",
                    getDocumentLocator());
                    // if invalid URI, ignore it
                }
            }
        }
    }

    private String resolvePrefix(String qname) throws SAXException {
        String prefix = "";
        int index = qname.indexOf(':');
        if (index > 0) prefix = qname.substring(0, index);
        String ns = getPrefixMapping(prefix);
        if (ns == null && !prefix.equals(""))
        throw new SAXParseException("Unknown namespace prefix " + prefix, getDocumentLocator());
        return ns;
    }

    private String getLocalName(String qname) {
        int index = qname.indexOf(':');
        if (index > 0) return qname.substring(index+1);
        else return qname;
    }

	private String getTableNamePart(String columnName, TableMappingImpl defaultTable) throws SAXException
	{
		String tablePart;
		int index = columnName.lastIndexOf('.');
		if (index != -1)
			return columnName.substring(0, index);
		else
		{
			TableMappingImpl tableMapping;
			if (defaultTable == null) {
				tableMapping = getCurrentTableMapping();
				if (tableMapping == null)
					throw new SAXParseException("No table name is specified in '" + columnName
						+ "' column name specification and no default table mapping was encountered.",
						getDocumentLocator());
			}
			else
				tableMapping = defaultTable;
				
			return tableMapping.getTableName();
		}
	}

	private String getTableNamePart(String columnName) throws SAXException
	{
		return getTableNamePart(columnName, null);
	}

    private String getColumnNamePart(String columnName) throws SAXException {
        int index = columnName.lastIndexOf('.');
        if (index == -1)
            return columnName;  // SR 18/12/2000 : tolerate column specification without table.
        else
            return columnName.substring(index+1);
    }

    private String checkColumnName(String columnName) throws SAXException {
        return checkColumnName(columnName, null);
    }

    private String checkColumnName(String columnName,TableMappingImpl tm) throws SAXException {
        String tableName = getTableNamePart(columnName);
        columnName = getColumnNamePart(columnName);
        if (tm == null) tm = findMappingInScope(tableName);
        else if (!tm.getTableName().equals(tableName))
        throw new SAXParseException("Column mapping cannot be defined here for table "
        +tableName, getDocumentLocator());
        if (tm == null)
        throw new SAXParseException("Table "+tableName+" is not in scope", getDocumentLocator());
        TableMetaData meta = tm.getMetaData();
        if (meta != null && meta.getColumnMetaData(columnName) == null)
        throw new SAXParseException("Column "+columnName+" is not a column of table "
        +tableName, getDocumentLocator());
        return columnName;
    }

    private ColumnMapping checkColumnRef(
        String columnName,
        TableMappingImpl tm,
        List refChildTableMappings)
        throws SAXException {
        
        String tableName = getTableNamePart(columnName);
        columnName = getColumnNamePart(columnName);
        // look in inner mapping first
		if (tm == null && refChildTableMappings != null) {
			int len = refChildTableMappings.size();
			TableMappingImpl myTm = null;
			for (int i = 0; i < len; i++) {
				myTm = (TableMappingImpl)refChildTableMappings.get(i);
				if (myTm.getTableName().equals(tableName))
					tm = myTm;
			}
		}
		// then in scope
		if (tm == null)
			tm = findMappingInScope(tableName);
        if (tm == null)
            throw new SAXParseException(
                "Table " + tableName + " is not in scope",
                getDocumentLocator());
		else if (!tm.getTableName().equals(tableName))
			throw new SAXParseException(
				"Column mapping does not refer to table " + tableName,
				getDocumentLocator());

        ColumnMapping cm = tm.getColumnMapping(columnName);
        if (cm == null)
            throw new SAXParseException(
                "Unknown column " + columnName + " in table " + tableName,
                getDocumentLocator());
        return cm;
    }

    private int getAction(Attributes atts) throws SAXException
    {
        String actionName = atts.getValue("", ACTION_ATTR);
        int action = INSERT;
        if (actionName != null)
        {
            if (actionName.equals(INSERT_VALUE)) action = INSERT;
            else if (actionName.equals(CHECK_VALUE)) action = CHECK;
            else if (actionName.equals(UPDATE_VALUE)) action = UPDATE;
            else if (actionName.equals(SELECT_VALUE)) action = SELECT;
            else
            throw new SAXParseException("Illegal action value: "+actionName,
                getDocumentLocator());
        }
        return action;
    }
    private int getBatchSize(Attributes atts) throws SAXException
    {
        int batchSize = RepositoryProperties.getIntProperty(RepositoryConstants.CONF_USER_BATCHSIZE);
        String size = atts.getValue("", BATCH_SIZE_ATTR);
        if (size != null)
            batchSize = Integer.parseInt(size);
        return batchSize;
    }

    /** Process key and select attributes for a column mapping building
     * the role property for column mapping.
     */
    private int getRole(Attributes atts,int def) throws SAXException
    {
        int role = def;
        String inKey = atts.getValue("", IN_KEY_ATTR);
        if (inKey != null)
        {
            if (inKey.equals(TRUE_VALUE) || inKey.equals(TRUE_NUM_VALUE)) role |= KEY_ROLE;
            else if (inKey.equals(FALSE_VALUE) || inKey.equals(FALSE_NUM_VALUE)) role &= ~KEY_ROLE;
            else
            throw new SAXParseException("Illegal value: "+inKey+" in inKey attribute", getDocumentLocator());
        }
        String inSelect = atts.getValue("", IN_SELECT_ATTR);
        if (inSelect != null)
        {
            if (inSelect.equals(TRUE_VALUE) || inSelect.equals(TRUE_NUM_VALUE))
                role |= SELECT_ROLE;
            else if (inSelect.equals(FALSE_VALUE) || inSelect.equals(FALSE_NUM_VALUE))
                role &= ~SELECT_ROLE;
            else
                throw new SAXParseException("Illegal value: " + inSelect + " in inSelect attribute",
                    getDocumentLocator());
        }
        return role;
    }

    private ElementDeclaration getElementDeclaration(String elementName)
    throws SAXException
    {
        if (elementName == null)
        throw new SAXParseException("Missing name attribute in element mapping declaration",
            getDocumentLocator());
        /* Look in the ElementDeclaration stack if present*/
        String namespace = resolvePrefix(elementName);
        String localName = getLocalName(elementName);
        ElementDeclaration decl = null;
        /* if the element corresponds to a complex type, look for the element in the complex type */
        if (context.isEmpty() && (currentType != null))
            decl = currentType.getElementDeclaration(namespace, localName);
        if (decl == null)
            decl = context.getElementDeclaration(namespace, localName);
        if (decl == null)
        throw new SAXParseException("Unknown element '"+elementName
        +"' in element mapping declaration", getDocumentLocator());
        return decl;
    }

    private AttributeDeclaration getAttributeDeclaration(String attrName) throws SAXException
    {
        if (attrName == null)
        throw new SAXParseException("Missing name attribute in attribute mapping declaration",
            getDocumentLocator());
        String namespace = resolvePrefix(attrName);
        String localName = getLocalName(attrName);
        AttributeDeclaration decl = null;
        if (context.isEmpty() && (currentType != null))
            decl = currentType.getAttributeDeclaration(namespace, localName);
        if (decl == null)
            decl = context.getAttributeDeclaration(namespace, localName);
        if (decl == null)
        throw new SAXParseException("Unknown attribute '"+attrName
        +"' in attribute mapping declaration", getDocumentLocator());
        return decl;
    }

    private Generator getUserGenerator(String genClass) throws SAXException
    {
        try {
            Object gen = Class.forName(genClass, true, classLoader).newInstance();
            if (gen instanceof UserGenerator)
                return (Generator)gen;
            else
                throw new SAXParseException("Class '" + genClass
                + "' is not an UserGenerator instance", getDocumentLocator());
        }
        catch (ClassNotFoundException ex1) {
            throw new SAXParseException("Generator class '"+genClass+"' not found",
                getDocumentLocator(), ex1);
        }
        catch (IllegalAccessException ex2) {
            throw new SAXParseException("Could not access constructor in generator class "
            + genClass, getDocumentLocator(), ex2);
        }
        catch (InstantiationException ex3) {
            throw new SAXParseException("Instantiation of generator class '"
            + genClass + "' failed", getDocumentLocator(), ex3);
        }
        catch (ClassCastException ex4) {
            throw new SAXParseException("Generator class '"+genClass
            +"' does not implement the interface org.xquark.xml.mapping.Generator",
            getDocumentLocator(), ex4);
        } catch (Exception ex) {
            throw new SAXParseException("Failed to load generator class " + genClass
            , getDocumentLocator(), ex);
        }
    }

    private Generator getValueGenerator() throws SAXException
    {
        SchemaComponent comp = getCurrentTableMapping().getSchemaComponent();
        return getValueGenerator(comp);
    }
    
    private Generator getValueGenerator(SchemaComponent comp) throws SAXException
    {
        Type type;
        if (comp instanceof Declaration)
            type = ((Declaration)comp).getType();
        else
            type = (Type)comp;
        if (type instanceof ComplexType)
        {
            int contentType = ((ComplexType)type).getContentType();
            if ((contentType == SchemaConstants.ELEMENT_ONLY)
                || (contentType == SchemaConstants.EMPTY))
                throw new SAXParseException(
                "Cannot reference $NodeValue in column mapping within a type mapping without character data.",
                getDocumentLocator());
        }
        SimpleType simple = type.getValueType();
        if (simple != null)
            return new ValueGenerator();
        
        throw new SAXParseException(
        "Cannot reference $NodeValue in complex element mapping", getDocumentLocator());
    }
    
    private Generator getSystemVariableGenerator(String source) throws SAXException
    {
        if (source.equals(VALUE_VALUE))
            return getValueGenerator();
        else
        {
            Generator generator = SystemVariableGenerator.createGenerator(source);
            if (generator == null)
                throw new SAXParseException("Unknown system variable " + source
                + " in column mapping declaration", getDocumentLocator());
            return generator;
        }
    }

	/* used for complexType table mapping only */
    private TableMappingImpl buildTableMapping(Attributes atts) throws SAXException
    {
        String typeName = atts.getValue("", TYPE_ATTR);
        if (typeName == null)
            throw new SAXParseException("Missing type attribute in type mapping specification",
                getDocumentLocator());
        String namespace = resolvePrefix(typeName);
        Type type = context.createTypeMapping(namespace, getLocalName(typeName)); // Update the schema context (create a root node for type)
        
        if (type == null)
        	throw new SAXParseException("Unknown type " + typeName
                + " in mapping declaration", getDocumentLocator());
        String name = atts.getValue("", NAME_ATTR);
        if (name == null || name.length() == 0)
        	throw new SAXParseException("Missing name in mapping declaration",
                    getDocumentLocator());
        TableMappingImpl res = buildTableMapping(type, atts);
        namedMappings.put(name, res);
        return res;
    }

    private TableMappingImpl buildTableMapping(SchemaComponent comp,Attributes atts) throws SAXException
    {
        TableMappingImpl result;
        String tableName = atts.getValue("", TABLE_ATTR);
        if (tableName == null || tableName.length() == 0)
            throw new SAXParseException("Missing table name in mapping declaration",
                getDocumentLocator());
        int action = getAction(atts);
        result = new TableMappingImpl(colMapping, comp, tableName, conn, action,
        generate, getBatchSize(atts), userTablesOnly);
        context.register(result);
        return result;
    }

    /**
     * Process generator column mapping object creation.
     */
    private void buildGeneratorColumnMapping(Attributes atts) throws SAXException
    {
        TableMappingImpl tableMapping = getCurrentTableMapping();
        String columnName = checkColumnName(atts.getValue("", COLUMN_ATTR),
        tableMapping);
        Generator generator = null;
        String ref = atts.getValue("", REF_ATTR);
        String source = atts.getValue("", VAR_ATTR);
        String genClass = atts.getValue("", METHOD_ATTR);
        // NIY       String XpathRef = atts.getValue("", XPATH_ATTR);

        int defaultRole = 0;
        if (ref != null)
        {
            if (source != null || genClass != null)
            throw new SAXParseException("Can have only one of " + REF_ATTR + ", "
                                    + VAR_ATTR + " and " + METHOD_ATTR
                                    + " attributes in generator declaration ("
                                    + tableMapping.getTableName() + "." + columnName + ").",
                                    getDocumentLocator());
			peekScope().addPendingRef(columnName, ref);		// processed when getting out of the surrounding element
//            generator = getColumnRefGenerator(ref);
            defaultRole = 0;
        }
        else if (source != null)
        {
            if (genClass != null)
            throw new SAXParseException("Can have only one of "
                                    + VAR_ATTR + " and " + METHOD_ATTR
                                    + " attributes in generator declaration ("
                                    + tableMapping.getTableName() + "." + columnName + ").",
                                    getDocumentLocator());
            generator = getSystemVariableGenerator(source);
            defaultRole = KEY_ROLE;
        }
        else if (genClass != null)
        {
            generator = getUserGenerator(genClass);
            defaultRole = KEY_ROLE;
        }
        else
            throw new SAXParseException("One of " + REF_ATTR + ", " + VAR_ATTR + " and " + METHOD_ATTR
                                    + " attributes must be specified in column mapping declaration ("
                                    + tableMapping.getTableName() + "." + columnName + ").",
                                    getDocumentLocator());
        int role = getRole(atts, defaultRole);
        // catch the value variable generator and make it a column mapping
        // of the element carrying the table mapping
        // TO IMPROVE : due to the language syntax, to fix the mapping of a leaf element with attributes...
        if (VALUE_VALUE.equals(source))            
        {
            Type type = null;
            if (tableMapping.getSchemaComponent() instanceof Type)
                type = (Type)tableMapping.getSchemaComponent();
            else     // declaration
                type = ((Declaration)tableMapping.getSchemaComponent()).getType();
            if ((type instanceof SimpleType) || ((type instanceof ComplexType)
                && (((ComplexType)type).getContentType() == SchemaConstants.TEXT_ONLY)))
            {
                context.register(new ColumnMappingImpl(tableMapping,
                            tableMapping.getSchemaComponent(),
                            1,
                            columnName,
                            getRole(atts, SELECT_ROLE),
                            generator,
                            messageHandler));
                return;
            }
            else
                throw new SAXParseException("System variable" + VALUE_VALUE
                        + " cannot be used for a complex content element ("
                        + tableMapping.getTableName() + "." + columnName + ").",
                        getDocumentLocator());
        }
        // in any other case       
        new ColumnMappingImpl(tableMapping, null, 1, columnName, role, generator, messageHandler);	// SR 14/05/2001
    }

    
    /** This method browse element hierarchy in schema to determine
     * the occurence bounds within a given ancestor.
     * @param elem element which occurence bounds are
     * to be calculated
     * @param ancestor scope where occurence is searched
     * @param occurrences an int[2] array for collecting
     * result. {min_occurs, max_occurs}
     * @return the child element declaration of the ancestor.
     */
    private ElementDeclaration computeHierarchyInfo(SchemaComponent ancestor,int[] occurrences)
    {
        ElementDeclaration child = null;
        SchemaComponent parent, current = null;
        ComplexType type;
        int[] occurrenceCount;
        long mult;
        Iterator it = context.iterator();
        if (it.hasNext())  // if we're processing an attribute elem is null and occurences is filled with atttribute occurence
            current = (ElementDeclaration)it.next();
        while (it.hasNext() && current != ancestor)
        {
            parent = (SchemaComponent)it.next();
            if (parent instanceof Declaration)
                type = (ComplexType)((Declaration)parent).getType();
            else
                type = (ComplexType)parent;
            
            occurrenceCount = type.getDeclarationOccurrence
                                (
                                current.getNamespace(),
                                current.getName()
                                );
            for (int i = 0; i <= 1; i++)
            {
                mult = (long)occurrences[i] * (long)occurrenceCount[i];
                occurrences[i] = (mult <= Integer.MAX_VALUE ? (int)mult : Integer.MAX_VALUE);
            }

            
            child = (ElementDeclaration)current;
            current = parent;
        }
        return child;
    }
    
    private String buildElementColumnMapping(String columnName,Attributes atts, ElementDeclaration decl) throws SAXException
    {
        String tableName = getTableNamePart(columnName);
        columnName = checkColumnName(columnName);
// TO REMOVE now passed as a parameter      ElementDeclaration decl =
//        getElementDeclaration(atts.getValue("", NAME_ATTR));
        TableMappingImpl tableMapping = findMappingInScope(tableName);
        int[] occurrences = { 1, 1 };
        ElementDeclaration ancestorChild =
        computeHierarchyInfo(tableMapping.getSchemaComponent(), occurrences);

        if (occurrences[1] > 1)
            throw new SAXParseException("Multivalued element " + decl
                + " cannot be mapped onto a column", getDocumentLocator());
        if (ancestorChild != null)
        tableMapping.addMappedChild(ancestorChild);
        String ref = atts.getValue("", REF_ATTR);
        
        Generator generator = null;
        if (ref == null)
            generator = getValueGenerator(decl);
        int role = getRole(atts, SELECT_ROLE);
        context.register(new ColumnMappingImpl(tableMapping, decl, occurrences[0],
            columnName, role, generator, messageHandler));
        return ref;
    }

	/** Ref generators are created here (should it be a real generator or a
	 * table-column mapping).
	 * @param columnName
	 * @param refTable
	 * @param refCol
	 * @throws SAXException
	 */
    private void completePendingRef(
        TableMappingImpl defaultTable,
        String columnName,
        TableMappingImpl refTable,
        String refCol,
        List refChildTableMappings,
        MappingNode originalContext
        )
        throws SAXException {
        Generator generator =
            new ColumnRefGenerator(checkColumnRef(refCol, refTable, refChildTableMappings));
        String tableName = getTableNamePart(columnName, defaultTable);
        columnName = checkColumnName(columnName);
        TableMappingImpl tableMapping = findMappingInScope(tableName);
        ColumnMappingImpl mapping =
            (ColumnMappingImpl) tableMapping.getColumnMapping(columnName);
        // Move the table-column mapping generator to the end (does nothing 4 ref generator)
		originalContext.moveColumnMappingToTheEnd(mapping);
        mapping.setGenerator(generator);
    }

    private void completePendingRefs(List refs) throws SAXException {
        if (refs != null) {
            int len = refs.size();
            RefItem refItem = null;
            for (int i = 0; i < len; i++) {
                refItem = (RefItem) refs.get(i);
                completePendingRef(
                    refItem.defautTableMapping,
                    refItem.columnName,
                    null,
                    refItem.ref,
                    refItem.childrenTableMappings,
                    refItem.contextNode
                    );
            }
        }
    }
	
    private TableMappingImpl completeTableMappingRef(
        Declaration decl,
        String columnName,
        String pendingRef,
        Attributes atts)
        throws SAXException {
        TableMappingImpl tableMapping = checkTableMappingRef(decl, atts);
        if (tableMapping == null)
            return null;
        completePendingRef(null, columnName, tableMapping, pendingRef, null, context.getCurrentNode());
        return tableMapping;
    }

    private TableMappingImpl checkTableMappingRef(Declaration decl, Attributes atts)
    throws SAXException
    {
        String map = atts.getValue("", MAP_ATTR);
        if (map == null) return null;
        TableMappingImpl tableMapping = (TableMappingImpl)namedMappings.get(map);
        if (tableMapping == null)
            throw new SAXParseException("Unknown named mapping for name '" + map
            + "'. Check the named table mapping is defined earlier in the mapping file.",
            getDocumentLocator());
        if (tableMapping.getSchemaType() != decl.getType())
            throw new SAXParseException("The mapping specified for name '" + map
            + "' does not match the element or attribute type.",
            getDocumentLocator());
        return tableMapping;
    }

    private String buildAttributeColumnMapping(String columnName,Attributes atts) throws SAXException
    {
        String tableName = getTableNamePart(columnName);
        columnName = checkColumnName(columnName);
        AttributeDeclaration decl =
        getAttributeDeclaration(atts.getValue("", NAME_ATTR));
        TableMappingImpl tableMapping = findMappingInScope(tableName);
        int[] occurrences = new int[2];
        switch (decl.getUse())
        {
            case SchemaConstants.REQUIRED:
                occurrences[0] = 1;
                occurrences[1] = 1;
                break;
            case SchemaConstants.OPTIONAL:
                occurrences[0] = 0;
                occurrences[1] = 1;
                break;
            case SchemaConstants.PROHIBITED:
                occurrences[0] = 0;
                occurrences[1] = 0;
                break;
        };
        ElementDeclaration ancestorChild =  computeHierarchyInfo(tableMapping.getSchemaComponent(), occurrences);

        if (occurrences[1] > 1)
            throw new SAXParseException("Multivalued attribute " + decl
            + " cannot be mapped onto a column", getDocumentLocator());
        if (ancestorChild != null)
        tableMapping.addMappedChild(ancestorChild);
        String ref = atts.getValue("", REF_ATTR);
        Generator generator = null;
        if (ref == null)
            generator = getValueGenerator(decl);
        int role = getRole(atts, SELECT_ROLE);
        context.register(new ColumnMappingImpl(tableMapping, decl, occurrences[0],
        columnName, role, generator, messageHandler));
        return ref;
    }

    private ElementDeclaration startElementMapping(Attributes atts) throws SAXException
	{
        ElementDeclaration decl =
        getElementDeclaration(atts.getValue("", NAME_ATTR));
        if ((atts.getValue("", REF_ATTR) != null ||
        atts.getValue("", IN_SELECT_ATTR) != null ||
        atts.getValue("", IN_KEY_ATTR) != null) &&
        atts.getValue("", COLUMN_ATTR) == null)
            throw new SAXParseException(
            "ref, inKey and inSelect attributes are allowed in element mapping declaration only when the column attribute is present.",
            getDocumentLocator());
        return decl;
    }
    private AttributeDeclaration startAttributeMapping(Attributes atts) throws SAXException {
        AttributeDeclaration decl =
        getAttributeDeclaration(atts.getValue("", NAME_ATTR));
        if ((atts.getValue("", COLUMN_ATTR) != null ||
        atts.getValue("", MAP_ATTR) != null ||
        atts.getValue("", IN_SELECT_ATTR) != null ||
        atts.getValue("", IN_KEY_ATTR) != null) &&
        atts.getValue("", REF_ATTR) == null)  // SR 14/11/2000
            throw new SAXParseException(
            "Only name attribute is allowed in attribute table mapping declaration when no column mapping reference is specified.",
            getDocumentLocator());
       return decl;
    }

    /** Process an element from the mapping file.
     * Used for root level (MAPPING element) & first level (ELEMENT or MAP elements) (TO BE IMPROVED).
     * Replaced by the corresponding handler at the deeper levels.
     */
    class MappingHandler extends DefaultElementHandler
    {
		private static final String RCSRevision = "$Revision: 1.6 $";
		private static final String RCSName     = "$Name:  $";
        public ElementHandler startElement(String namespaceURI,String localName,Attributes atts)
        throws SAXException
        {
            if (isMappingNamespace(namespaceURI))
            {
                /* process the root MAPPING element (level 0) */
                if (localName.equals(MAPPING_TAG))
                {
                    initMapping(atts);
                    return this;    // Provide handler for level 1
                }
                /* process the MAP element (level 1 for Type Mapping) */
                else if (localName.equals(MAP_TAG))
                {
                    TableMappingImpl mapping = buildTableMapping(atts);
                    currentType = (Type)mapping.getSchemaComponent();
                    pushScope(mapping);    // put the last Table mapping encountered as the first to consider (push)
                    return new TableMappingHandler();   // Provide handler for next level
                }
                /* process the ELEMENT element (level 1) containing only a MAP element (because of the level) */
                else if (localName.equals(ELEMENT_TAG))
                {
                    ElementDeclaration decl = startElementMapping(atts);    // Retrieves the Element decl from the schema
                    context.createElementMapping(decl, (checkTableMappingRef(decl, atts) != null)); // Update the schema context

					/* Create scope */
					pushScope();
                    
                    return new ElementMappingHandler();   // Provide handler for next level
                }
            }
            notifyUnknownElement(namespaceURI, localName);
            return this;
        }

        public void endElement(String namespaceURI,String localName) throws SAXException {
            if (isMappingNamespace(namespaceURI))
            {
                if (localName.equals(MAP_TAG))
                {
                    getCurrentTableMapping().initialize(conn);   // Now Column mappings are created
                    // NOTE: Scope are now attached to the surrounding element, except for top-level mappings
                    popScope();
                    context.popType(); // Update the schema context removing the Element declaration from the schema
                    currentType = null;
                }
                else if (localName.equals(ELEMENT_TAG)) {
                    postProcessEnclosedTableMapping();
					context.pop(); // Update the schema context removing the Element declaration from the schema
                }
                // no op when getting out of MAPPING element.
            }
        }


    }

    /** Process an element from the mapping file when the element encountered
     * is not a MAP element, i.e., process ELEMENT, ATRIBUTE or GENERATOR elements.
     * If the mapping element is bound to a column mapping this object is
     * returned and recycled.
     * If the mapping element has got a content (ELEMENT or ATTRIBUTE) a new
     * is created.
     */
    class TableMappingHandler extends DefaultElementHandler
    {
		private static final String RCSRevision = "$Revision: 1.6 $";
		private static final String RCSName     = "$Name:  $";

        private ElementDeclaration eltDecl =null;
        private AttributeDeclaration attDecl =null;
        private String columnName =null;
        private boolean childExpected = false;

        public ElementHandler startElement(String namespaceURI,String localName,Attributes atts)
		throws SAXException
		{
			String pendingRef =null;
			childExpected = false;
			
            if (isMappingNamespace(namespaceURI))
            {
                /* Process the tag corresponding to a column mapping */
                if (localName.equals(GENERATOR_TAG))
                {
                    buildGeneratorColumnMapping(atts);
                    return this;
                }
                /* Process the tag corresponding to an ELEMENT element */
                else if (localName.equals(ELEMENT_TAG))
                {
                    // Starts a new element table mapping declaration, which should
                    // resolve pending ref if it is not null
                    eltDecl = startElementMapping(atts);    // Retrieves the Element decl from the schema
                    context.createElementMapping(eltDecl, (checkTableMappingRef(eltDecl, atts) != null)); // Update the schema context (Now done event for element with no table mapping, to build RepositoryMapping tree)
					
                    /* A COLUMN attribute exists : it is an empty element corresponding to a column mapping*/
                    columnName = atts.getValue("",COLUMN_ATTR);
                    RefItem refItem = null;
                    if (columnName != null)
                    {
                        // build column mapping, returns pending ref if mapping
                        // refers to a foreign table
                        pendingRef = buildElementColumnMapping(columnName, atts, eltDecl);
                        if (pendingRef == null)
                        {
                            // Column mapping complete
                            return this;
                        }
                        else
                        {
                            // check if mapping is a reference to an existing type mapping,
                            // which should resolve pending ref.
                            // Returns null if not a reference
                            TableMapping mapping = completeTableMappingRef(eltDecl, columnName, pendingRef, atts);
                            if (mapping == null)
                            	refItem = peekScope().addPendingRef(columnName, pendingRef);
                            else
                                return this;
                        }
                    }
                    
					/* Create scope */
					childExpected = true;
					pushScope();
                    
                    if (refItem != null)
                    	refItem.childrenTableMappings = peekScope().getTableMappings();
                    
                    /* This ELEMENT element has a content. A new ElementMappingHandler is created */
                    return new ElementMappingHandler();   // Provide handler for the content elements
                }
                /* Process the tag corresponding to an ATTRIBUTE element */
                else if (localName.equals(ATTRIBUTE_TAG))
                {
                    /* A COLUMN attribute exists : it is an empty element corresponding to a column mapping*/
                    columnName = atts.getValue("",COLUMN_ATTR);
                    if (columnName != null)
                    {
                        // build column mapping, returns pending ref if mapping
                        // refers to a foreign table
                        pendingRef = buildAttributeColumnMapping(columnName, atts);
                        if (pendingRef == null)
                        {
                            // Column mapping complete
                            return this;
                        }
                        else
                        {
                            // check if mapping is a reference to an existing type mapping,
                            // which should resolve pending ref.
                            // Returns null if not a reference
                            TableMapping mapping =  completeTableMappingRef(
                                getAttributeDeclaration(atts.getValue("", NAME_ATTR)),
                                columnName, pendingRef, atts);
							if (mapping == null)
								peekScope().addPendingRef(columnName, pendingRef);
							else
								return this;
                        }
                    }

					/* Create scope */
					childExpected = true;
					pushScope();
                    
                    /* This ATTRIBUTE element has a content. A new AttributeMappingHandler is created */
                    // Starts a new element table mapping declaration
                    attDecl = startAttributeMapping(atts);    // Retrieves the Attribute decl from the schema
                    return new AttributeMappingHandler(attDecl);   // Provide handler for the MAP element (only content possible for ATTRIBUTE element)
                }
            }
            notifyUnknownElement(namespaceURI, localName);
            return this;
        }

        public void endElement(String namespaceURI,String localName) throws SAXException {
            if (isMappingNamespace(namespaceURI))
            {
                if (localName.equals(ELEMENT_TAG) || localName.equals(ATTRIBUTE_TAG))
                {
					if (childExpected)
						postProcessEnclosedTableMapping();
					if (eltDecl != null)
                    {
                        eltDecl = null;
                        context.pop();
                    }
                }
            }
        }
    }
	
	private void postProcessEnclosedTableMapping() throws SAXException
	{
		ScopeItem scopeItem = peekScope();
		
		// Complete pending refs every table mapping encountered
		completePendingRefs(scopeItem.getPendingRefs());

		// Initialize table mappings (Now Column mappings are created)
		Iterator it = scopeItem.getTableMappings().iterator();
                	
		while (it.hasNext()) {
			((TableMappingImpl) it.next()).initialize(conn);
		}
		popScope();
	}
	
    /** Processes the content of an ELEMENT element.
     * Behaves like a TableMappingHandler except if the current mapping element
     * is a MAP one. A new TableMappingHandler is then created.
     */
    class ElementMappingHandler extends TableMappingHandler {
		private static final String RCSRevision = "$Revision: 1.6 $";
		private static final String RCSName     = "$Name:  $";
        public ElementHandler startElement(String namespaceURI,String localName,Attributes atts) throws SAXException {
            if (isMappingNamespace(namespaceURI))
            {
                if (localName.equals(MAP_TAG))
                {
                    addMapping2Scope(buildTableMapping(context.top(), atts));
                    return new TableMappingHandler();
                }
                else
                return super.startElement(namespaceURI, localName, atts);
            }
            notifyUnknownElement(namespaceURI, localName);
            return this;
        }

    }

    /** Processes the content of an ATTRIBUTE element (i.e., MAP element).
     * A new TableMappingHandler is created.
     */
    class AttributeMappingHandler extends DefaultElementHandler {
		private static final String RCSRevision = "$Revision: 1.6 $";
		private static final String RCSName     = "$Name:  $";
        private AttributeDeclaration attDecl;

        AttributeMappingHandler(AttributeDeclaration decl) {
            attDecl = decl;
        }

        public ElementHandler startElement(String namespaceURI,String localName,Attributes atts) throws SAXException {
            if (isMappingNamespace(namespaceURI))
            {
                if (localName.equals(MAP_TAG))
                {
					addMapping2Scope(buildTableMapping(attDecl, atts));
                    return new TableMappingHandler();
                }
            }
            notifyUnknownElement(namespaceURI, localName);
            return this;
        }
    }
    
	private TableMappingImpl getCurrentTableMapping() {
		for (int i = scope.size() - 1; i >= 0; i--) {
			TableMappingImpl tm = ((ScopeItem)scope.get(i)).getLastInserted();
			if (tm != null)
				return tm;
		}
		return null;
	}
    
	private TableMappingImpl findMappingInScope(String tableName) {
		for (int i = scope.size() - 1; i >= 0; i--) {
			List tms = ((ScopeItem)scope.get(i)).getTableMappings();
			for (int j = 0; j < tms.size(); j++) {
				TableMappingImpl tm = (TableMappingImpl)tms.get(j);
				if (tm.getTableName().equals(tableName)) return tm;
			}
		}
		return null;
	}

	private void addMapping2Scope(TableMappingImpl tm) {
		peekScope().addTableMapping(tm);
	}

	private void pushScope() {
		scope.add(new ScopeItem());
	}
	
	private void pushScope(TableMappingImpl tm) {
		scope.add(new ScopeItem(tm));
	}
	
	private ScopeItem peekScope() {
		return (ScopeItem) scope.get(scope.size() - 1);
	}
	
	private ScopeItem popScope() {
		return (ScopeItem) scope.remove(scope.size() - 1);
	}
	
    private class ScopeItem {
    	/** A map between table mappings and a list of pending re:f {tm, [ref1,ref2,..]} */
    	private List tmList = new ArrayList(1);
    	private List refList = null;
    	
		ScopeItem() {
		}
		
    	ScopeItem(TableMappingImpl tm) {
    		this();
    		addTableMapping(tm);
    	}
    	
		void addTableMapping(TableMappingImpl tm) {
			tmList.add(tm);
		}
		
		List getPendingRefs() {
			return refList;
		}
		
		List getTableMappings() {
			return tmList;
		}
		
		TableMappingImpl getLastInserted() {
			TableMappingImpl ret = null;
			if (!tmList.isEmpty())
				ret = (TableMappingImpl)tmList.get(tmList.size() - 1);
			return ret;
		}
		
		RefItem addPendingRef(String column, String pendingRef) {
			RefItem ret = null;
			if (refList == null)
				refList = new ArrayList(2);
			ret = new RefItem(column, pendingRef, getLastInserted());
			refList.add(ret);
			return ret;
		}
    }
    
    private class RefItem {
    	String columnName = null;
    	String ref = null;
    	TableMappingImpl defautTableMapping =  null;
    	List childrenTableMappings = null;
		MappingNode contextNode = null;
    	
        RefItem(String columnName, String ref, TableMappingImpl defautTableMapping) {
            this.columnName = columnName;
            this.ref = ref;
            this.defautTableMapping = defautTableMapping;
			contextNode = context.getCurrentNode();
        }
    }
}






